"""Models and functions used for configuration definitions."""
from __future__ import annotations

from collections import OrderedDict
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Mapping,
    Optional,
)
from typing import OrderedDict as OrderedDictType
from typing import (
    Text,
    Type,
    TypeVar,
    Union,
    cast,
)

from ywh2bt.core.configuration.error import (
    AttributesError,
    BaseAttributeError,
    InvalidAttributeError,
    MissingAttributeError,
    UnsupportedAttributeError,
)
from ywh2bt.core.configuration.exportable import Exportable
from ywh2bt.core.configuration.validatable import Validatable
from ywh2bt.core.configuration.validator import (
    ValidatorError,
    ValidatorProtocol,
)


T = TypeVar("T", covariant=True)
V = TypeVar("V", covariant=True)


class Attribute(Generic[T]):
    """
    A class representing an attribute of a configuration.

    Examples:
        class MyContainer(AttributesContainer):
            my_string_attribute = Attribute.create(value_type='str')

        instance = MyContainer()
        instance.my_string_attribute = 'foo'
        print(instance.my_string_attribute)
    """

    _name: Optional[str] = None
    value_type: Type[T]
    secret: bool
    required: bool
    default: Optional[T]
    deprecated: bool
    short_description: Optional[Text] = None
    description: Optional[Text] = None
    validator: Optional[ValidatorProtocol[T]] = None

    def __init__(
        self,
        value_type: Type[T],
        short_description: Optional[Text] = None,
        description: Optional[Text] = None,
        default: Optional[T] = None,
        secret: bool = False,
        required: bool = False,
        deprecated: bool = False,
        validator: Optional[ValidatorProtocol[T]] = None,
    ) -> None:
        """
        Initialize the attribute.

        Args:
            value_type: a type expected for the value of the attribute
            short_description: a short description of the attribute
            description: a description of the attribute
            default: a default value for the attribute
            secret: a flag indicating if the attribute is secret (avoid representing/printing sensible data)
            required: a flag indicating if the attribute must have a non-None value
            deprecated: a flag indicating if the attribute is deprecated
            validator: a function validating the value of the attribute

        Raises:
            InvalidAttributeError: if the attribute could not be initialized
        """
        self.value_type = value_type
        self.short_description = short_description
        self.description = description
        self.secret = secret
        self.required = required
        self.deprecated = deprecated
        self.default = default
        self.validator = validator
        if default is not None:
            self._ensure_default_value_type()
            if validator:
                try:
                    validator(
                        value=default,
                    )
                except ValidatorError as e:
                    raise InvalidAttributeError(
                        message=f"Validation error for default value ({repr(default)}): {e}",
                        context=self,
                    )

    @classmethod
    def create(
        cls,
        value_type: Type[V],
        short_description: Optional[Text] = None,
        description: Optional[Text] = None,
        default: Optional[V] = None,
        secret: bool = False,
        required: bool = False,
        deprecated: bool = False,
        validator: Optional[ValidatorProtocol[V]] = None,
    ) -> Attribute[V]:
        """
        Create a new attribute.

        Must only be used in the definition of classes extending `AttributesContainer`.

        Args:
            value_type: a type expected for the value of the attribute
            short_description: a short description of the attribute
            description: a description of the attribute
            default: a default value for the attribute
            secret: a flag indicating if the attribute is secret (avoid representing/printing sensible data)
            required: a flag indicating if the attribute must have a non-None value
            deprecated: a flag indicating if the attribute is deprecated
            validator: a function validating the value of the attribute

        Returns:
            The newly created attribute
        """
        return Attribute[V](
            value_type=value_type,
            short_description=short_description,
            description=description,
            default=default,
            secret=secret,
            required=required,
            deprecated=deprecated,
            validator=validator,
        )

    def validate(
        self,
        value: Any,
    ) -> None:
        """
        Validate a possible value for the attribute.

        Args:
            value: a value to be validated
        """
        if value is None:
            self._ensure_required_has_default()
            value = self.default
        if value is not None:
            self._ensure_value_type(value=value)
            self._ensure_validity(value=value)

    def __set__(
        self,
        instance: AttributesContainer,
        value: Any,
    ) -> None:
        """
        Set the value of a descriptor.

        Is called to set the attribute on an instance `instance` of the owner class to a new value, `value`.

        Args:
            instance: an instance of the owner class
            value: a value

        Returns:
            None
        """
        if value:
            if isinstance(value, Dict):
                value = self.value_type(**value)
            elif isinstance(value, List):
                value = self.value_type(value)  # type: ignore
        return instance.set_attribute_value(
            attribute=self,
            value=value,
        )

    def __get__(
        self,
        instance: AttributesContainer,
        owner: Type[AttributesContainer],
    ) -> Optional[T]:
        """
        Get a computed value of a descriptor.

        Is called to get the attribute of the owner class (class attribute access)
        or of an instance of that class (instance attribute access).

        Args:
            instance: instance that the attribute was accessed through
            owner: owner class

        Returns:
            The computed value
        """
        return (
            cast(
                Optional[T],
                instance.get_attribute_value(
                    attribute=self,
                ),
            )
            if instance is not None
            else self
        )

    def __set_name__(
        self,
        owner: Type[AttributesContainer],
        name: str,
    ) -> None:
        """
        Is called at the time the owning class `owner` is created.

        Args:
            owner: owner class
            name: a descriptor name
        """
        self._name = name

    @property
    def name(self) -> str:
        """
        Get the name of the attribute.

        Returns:
            The name of the attribute
        """
        if self._name is not None:
            return self._name
        return f"Unnamed attribute #{hash(self):x}"

    def as_repr(
        self,
        value: Optional[T],
    ) -> str:
        """
        Get a safe string representation of the value depending on the `secret` flag.

        Args:
            value: a value to be represented

        Returns:
            The safe string representation
        """
        if value is not None and self.secret:
            return f"<secret-{self.value_type.__name__} '***'>"
        return repr(value)

    def __repr__(self) -> str:
        """
        Is called by the `repr()` built-in function to compute the "official" string representation.

        Returns:
            The string representation
        """
        items = [
            f"name={repr(self.name)}",
            f"value_type={repr(self.value_type)}",
            f"required={repr(self.required)}",
            f"secret={repr(self.secret)}",
        ]
        if self.short_description:
            items.append(f"short_description={repr(self.short_description)}")
        if self.description:
            items.append(f"description={repr(self.description)}")
        if self.default is not None:
            items.append(f"default={self.as_repr(self.default)}")

        joined_items = ", ".join(items)
        return f"{self.__class__.__name__}({joined_items})"

    def _ensure_default_value_type(self) -> None:
        if not isinstance(self.default, self.value_type):
            default_type = type(self.default)
            safe_default_repr = self.as_repr(self.default)
            message = f"Expecting {self.value_type} for default value ; got {default_type} ({safe_default_repr})."
            raise InvalidAttributeError(
                message=message,
                context=self,
            )

    def _ensure_required_has_default(self) -> None:
        if self.default is None:
            if self.required:
                raise MissingAttributeError(
                    message=f"Expecting value for required attribute {repr(self.name)} with no default",
                    context=self,
                )

    def _ensure_value_type(
        self,
        value: Any,
    ) -> None:
        if not isinstance(value, self.value_type):
            name_repr = repr(self.name)
            if self.short_description:
                name_repr = f"{self.short_description} ; {name_repr}"
            value_type = type(value)
            safe_value_repr = self.as_repr(value)
            message = (
                f"Wrong value type for {name_repr} = {safe_value_repr}: expecting {self.value_type} got {value_type}"
            )
            raise InvalidAttributeError(
                message=message,
                context=self,
            )

    def _ensure_validity(
        self,
        value: Any,
    ) -> None:
        if isinstance(value, Validatable):
            value.validate()
        if self.validator:
            self.apply_validator(
                validator=self.validator,
                value=value,
            )

    def apply_validator(
        self,
        validator: ValidatorProtocol[T],
        value: Any,
    ) -> None:
        """
        Apply the validator to the given value.

        Args:
            validator: a validator
            value: a value

        Raises:
            InvalidAttributeError: if the attribute value is invalid
        """
        try:
            validator(
                value=value,
            )
        except ValidatorError as e:
            name_repr = repr(self.name)
            if self.short_description:
                name_repr = f"{self.short_description} ; {name_repr}"
            safe_value_repr = self.as_repr(value)
            message = f"Validation error for {name_repr} = {safe_value_repr}: {e}"
            raise InvalidAttributeError(
                message=message,
                context=self,
            )


class AttributesContainer(
    Validatable,
    Exportable[Dict[str, Any]],
):
    """A class representing a container for all the attributes of a configuration."""

    _values: Dict[Attribute[Any], Any]
    _extra: Dict[str, Any]

    def __init__(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the container.

        Args:
            kwargs: keyword arguments
        """
        self._values = {}
        for _, value in self.get_attributes().items():
            if isinstance(value, Attribute):
                self._values[value] = None
        self._extra = kwargs or {}

    @classmethod
    def get_attributes(cls) -> Dict[str, Attribute[Any]]:
        """
        Get all created attributes.

        Returns:
            The attributes
        """
        attributes = OrderedDict()
        for name, value in vars(cls).items():
            if isinstance(value, Attribute):
                attributes[name] = value
        return attributes

    def set_attribute_value(
        self,
        attribute: Attribute[Any],
        value: Any,
    ) -> None:
        """
        Set the value of an attribute.

        Args:
            attribute: an attribute
            value: a value
        """
        self._values[attribute] = value

    def get_attribute_value(
        self,
        attribute: Attribute[Any],
    ) -> Any:
        """
        Get the value of an attribute.

        Args:
            attribute: an attribute

        Returns:
            The attribute value
        """
        value = self._values[attribute] if attribute in self._values else attribute.default
        return value if value is not None else attribute.default

    def validate(self) -> None:
        """
        Validate the container.

        Raises:
            AttributesError: if the container is not valid
        """
        errors: Dict[str, BaseAttributeError] = {}
        for attribute, value in self._values.items():
            try:
                attribute.validate(
                    value=value,
                )
            except BaseAttributeError as e:
                errors[attribute.name] = e
        for extra_key, _ in self._extra.items():
            errors[extra_key] = UnsupportedAttributeError(
                message="Unsupported attribute",
                context=self,
            )
        if errors:
            raise AttributesError(
                message="Validation failed",
                errors=errors,
                context=self,
            )

    def export(self) -> Dict[str, Any]:
        """
        Export the container values as a dictionary.

        Returns:
            The exported values
        """
        values: Dict[str, Any] = {}
        for attribute, value in self._values.items():
            if isinstance(value, Exportable):
                exported = value.export()
                if exported is not None:
                    values[attribute.name] = exported
            elif value is not None:
                values[attribute.name] = value
        values.update(self._extra)
        return values

    def __str__(self) -> str:
        """
        Get the string value.

        Is called by `str(object)` and the built-in functions `format()` and `print()`
        to compute the "informal" or nicely printable string representation.

        Returns:
            The string value
        """
        joined_items = ", ".join(
            [
                f"{attr.name}={attr.as_repr(value=self.get_attribute_value(attribute=attr))}"
                for attr, value in self._values.items()
            ]
        )
        return f"{self.__class__.__name__}({joined_items})"

    def __repr__(self) -> str:
        """
        Is called by the `repr()` built-in function to compute the "official" string representation.

        Returns:
            The string representation
        """
        return self.__str__()


T_AC = TypeVar("T_AC", bound=AttributesContainer)


class AttributesContainerList(
    List[T_AC],
    Validatable,
    Exportable[List[Any]],
):
    """A class representing a typed list of attributes container."""

    values_type: Type[T_AC]

    def __init__(
        self,
        values_type: Type[T_AC],
        items: Optional[List[T_AC]],
    ):
        """
        Initialize the list.

        Args:
            values_type: a type for the items of the list
            items: initial items to be added to the list
        """
        new_seq: List[T_AC] = []
        if items:
            for item in items:
                if isinstance(item, values_type):
                    new_seq.append(item)
                elif item is not None:
                    new_seq.append(values_type(**item))  # type: ignore
        super().__init__(new_seq)
        self.values_type = values_type

    def append(self, obj: Optional[T_AC]) -> None:
        """
        Append an item to the list.

        This also tries to cast the value to the relevant type if the value is not an instance of the expected type.

        Args:
            obj: a item to be appended
        """
        if obj is not None:
            super().append(
                obj if isinstance(obj, self.values_type) else self.values_type(**obj),  # type: ignore
            )

    def validate(self) -> None:
        """
        Validate the list.

        Raises:
            AttributesError: if the list is not valid
        """
        errors: Dict[str, BaseAttributeError] = {}
        for i, item in enumerate(self):
            if not isinstance(item, self.values_type):
                item_type = type(item)
                item_repr = repr(item)
                errors[str(i)] = InvalidAttributeError(
                    message=(f"Expecting {self.values_type} for item {i} ; " + f"got {item_type} ({item_repr})"),
                    context=self,
                )
            elif isinstance(item, Validatable):
                try:
                    item.validate()
                except BaseAttributeError as e:
                    errors[str(i)] = e
        if errors:
            raise AttributesError(
                message="Validation failed",
                errors=errors,
                context=self,
            )

    def export(self) -> List[Any]:
        """
        Export the list.

        Returns:
            The exported values
        """
        values: List[Any] = []
        for container in self:
            if isinstance(container, Exportable):
                values.append(container.export())
            elif container is not None:
                values.append(container)
        return values


KT = TypeVar("KT")
VT = TypeVar("VT")
EKT = TypeVar("EKT")
EVT = TypeVar("EVT")


class ExportableDict(OrderedDictType[KT, VT], Exportable[Dict[EKT, EVT]]):
    """A class describing an exportable dict."""

    def export(self) -> Dict[EKT, EVT]:
        """
        Export the dict.

        Returns:
            the exported dict
        """
        return dict(self)  # type: ignore

    def swap_key(
        self,
        old: KT,
        new: KT,
    ) -> None:
        """
        Swap item key.

        Args:
            old: an old key
            new: a new key
        """
        for _ in range(len(self)):
            k, v = self.popitem(False)
            self[new if old == k else k] = v


class AttributesContainerDict(
    ExportableDict[str, T_AC, str, Any],
    Validatable,
):
    """A class representing dictionary of attributes container."""

    values_type: Type[T_AC]

    def __init__(
        self,
        values_type: Type[T_AC],
        items: Mapping[str, Any],
    ) -> None:
        """
        Initialize the dictionary.

        Args:
            values_type: a type for the items of the dict
            items: initial items to be added to the dict

        Raises:
            AttributesError: if an item is invalid
        """
        self.values_type = values_type
        errors = {}
        new_items: Dict[str, T_AC] = {}
        for key, item in items.items():
            try:
                if isinstance(item, values_type):
                    new_items[key] = item
                elif item is not None:
                    new_items[key] = self._init_item(
                        item=item,
                    )
            except BaseAttributeError as e:
                errors[key] = e
        if errors:
            raise AttributesError(
                message="Invalid items",
                context=self,
                errors=errors,
            )
        super().__init__(new_items)

    def _init_item(
        self,
        item: Any,
    ) -> T_AC:
        return self.values_type(**item)

    def validate(self) -> None:
        """
        Validate the dictionary.

        Raises:
            AttributesError: if the list is not valid
        """
        errors: Dict[str, BaseAttributeError] = {}
        for key, container in self.items():
            try:
                container.validate()
            except BaseAttributeError as e:
                errors[key] = e
        if errors:
            raise AttributesError(
                message="Validation failed",
                errors=errors,
                context=self,
            )

    def export(self) -> Dict[str, Any]:
        """
        Export the dictionary.

        Returns:
            The exported values
        """
        values: Dict[str, Any] = {}
        for key, container in self.items():
            if isinstance(container, Exportable):
                values[key] = container.export()
            elif container is not None:
                values[key] = container
        return values


StrAttributeType = Union[Optional[str], Attribute[str]]
BoolAttributeType = Union[Optional[bool], Attribute[bool]]

LT = TypeVar("LT")
ET = TypeVar("ET")


class ExportableList(List[LT], Exportable[List[ET]]):
    """A list."""

    def export(self) -> List[ET]:
        """
        Export the list.

        Returns:
            the exported list
        """
        return cast(List[ET], self.copy())



"""Models used for the root configuration."""
from typing import (
    Any,
    Dict,
    Optional,
    Union,
    cast,
)

from ywh2bt.core.configuration.attribute import (
    Attribute,
    AttributesContainer,
)
from ywh2bt.core.configuration.error import AttributesError
from ywh2bt.core.configuration.tracker import (
    TrackerConfiguration,
    Trackers,
)
from ywh2bt.core.configuration.validator import not_empty_validator
from ywh2bt.core.configuration.yeswehack import (
    Programs,
    YesWeHackConfigurations,
)


BugtrackersAttributeType = Union[
    Dict[str, TrackerConfiguration],
    Optional[Trackers],
    Attribute[Trackers],
]
BugtrackersInitType = Union[
    Optional[Trackers],
    Optional[Dict[str, TrackerConfiguration]],
    Optional[Dict[str, Any]],
]
YesWeHackAttributeType = Union[
    Dict[str, Any],
    Optional[YesWeHackConfigurations],
    Attribute[YesWeHackConfigurations],
]
YesWeHackInitType = Union[
    Optional[YesWeHackConfigurations],
    Optional[Dict[str, Any]],
]


class RootConfiguration(AttributesContainer):
    """A root configuration."""

    bugtrackers: BugtrackersAttributeType = Attribute.create(
        value_type=Trackers,
        short_description="Trackers",
        description="Trackers to be synchronized with YesWeHack",
        required=True,
        validator=not_empty_validator,
    )
    yeswehack: YesWeHackAttributeType = Attribute.create(
        value_type=YesWeHackConfigurations,
        short_description="YesWeHack",
        required=True,
        validator=not_empty_validator,
    )

    def __init__(
        self,
        bugtrackers: BugtrackersInitType = None,
        yeswehack: YesWeHackInitType = None,
        **kwargs: Any,
    ):
        """
        Initialize the configuration.

        Args:
            bugtrackers: a configuration for the bugtrackers
            yeswehack: a configuration for YesWeHack
            kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.bugtrackers = bugtrackers
        self.yeswehack = yeswehack

    def validate(self) -> None:
        """
        Validate the configuration.

        Raises:
            AttributesError: if the configuration is invalid

        # noqa: DAR401
        # noqa: DAR402
        """
        super_error: Optional[AttributesError] = None
        try:
            super().validate()
        except AttributesError as e:
            super_error = e

        errors = {}
        if self.yeswehack:
            errors.update(
                self._ensure_declared_trackers(
                    trackers=cast(Trackers, self.bugtrackers or {}),
                    yeswehack=cast(YesWeHackConfigurations, self.yeswehack or {}),
                ),
            )

        if errors:
            if not super_error:
                super_error = AttributesError(
                    message="Validation error",
                    errors={},
                    context=self,
                )
            super_error.errors.update(errors)

        if super_error:
            raise super_error

    def _ensure_declared_trackers(
        self,
        trackers: Trackers,
        yeswehack: YesWeHackConfigurations,
    ) -> Dict[str, AttributesError]:
        errors = {}
        for yeswehack_name, yeswehack_config in yeswehack.items():
            config_errors = self._ensure_bugtrackers_name(
                trackers=trackers,
                programs=cast(Programs, yeswehack_config.programs or {}),
            )
            for key, error in config_errors.items():
                errors[f"yeswehack.{yeswehack_name}.{key}"] = error
        return errors

    def _ensure_bugtrackers_name(
        self,
        trackers: Trackers,
        programs: Programs,
    ) -> Dict[str, AttributesError]:
        errors = {}
        for i, program in enumerate(programs):
            if not program.bugtrackers_name:
                continue
            for j, bugtracker_name in enumerate(program.bugtrackers_name):
                if bugtracker_name not in trackers:
                    errors[f"programs[{i}].bugtrackers_name[{j}]"] = AttributesError(
                        message=f"Bugtracker not {repr(bugtracker_name)} declared in bugtrackers",
                        errors={},
                        context=self,
                    )
        return errors



"""Models and functions used for YesWeHack platform data access."""
from json import JSONDecodeError
from typing import (
    Any,
    Dict,
    List,
    Optional,
    cast,
)
from urllib.parse import urlsplit

import requests
from yeswehack.api import Attachment as YesWeHackRawApiAttachment
from yeswehack.api import Log as YesWeHackRawApiLog
from yeswehack.api import Report as YesWeHackRawApiReport
from yeswehack.api import YesWeHack as YesWeHackRawApiClient
from yeswehack.exceptions import APIError as YesWeHackRawAPiError

from ywh2bt.core.api.client import TestableApiClient
from ywh2bt.core.api.mapping import (
    MappingContext,
    map_raw_attachment,
    map_raw_logs,
    map_raw_report,
)
from ywh2bt.core.api.models.report import (
    Attachment,
    Log,
    Report,
)
from ywh2bt.core.configuration.yeswehack import YesWeHackConfiguration
from ywh2bt.core.exceptions import CoreException
from ywh2bt.version import __VERSION__


class YesWeHackApiClientError(CoreException):
    """A YesWeHack API client error."""


class YesWeHackApiClient(TestableApiClient):
    """A YesWeHack API client."""

    _configuration: YesWeHackConfiguration
    _raw_client: YesWeHackRawApiClient
    _yeswehack_domain: str
    _logged_in: bool

    def __init__(
        self,
        configuration: YesWeHackConfiguration,
    ):
        """
        Initialize self.

        Args:
            configuration: a YesWeHack configuration
        """
        self._configuration = configuration
        self._raw_client = self._build_raw_client()
        self._logged_in = False
        self._yeswehack_domain = self._extract_yeswehack_domain(
            url=cast(str, self._configuration.api_url),
        )

    def test(
        self,
    ) -> None:
        """Test the client."""
        self._ensure_login()

    def _build_raw_client(
        self,
    ) -> YesWeHackRawApiClient:
        configuration = self._configuration
        try:
            client = YesWeHackRawApiClient(
                api_url=self._normalize_api_url(
                    api_url=cast(str, configuration.api_url),
                ),
                pat=configuration.pat,
                verify=configuration.verify,
                lazy=True,
                headers={
                    "User-Agent": f"ywh2bt/{__VERSION__}",
                },
            )
        except (YesWeHackRawAPiError, requests.RequestException) as e:
            raise YesWeHackApiClientError("Unable to initialize YesWeHack API client") from e
        return client

    def _normalize_api_url(
        self,
        api_url: str,
    ) -> str:
        return api_url[:-1] if api_url[-1] == "/" else api_url

    def _extract_yeswehack_domain(
        self,
        url: str,
    ) -> str:
        result = urlsplit(url)
        # only keep 'xxx.yyy.zzz' from 'www.xxx.yyy.zzz'
        _, domain = result.netloc.split(".", 1)
        return domain

    def _ensure_login(
        self,
    ) -> None:
        if not self._logged_in:
            try:
                success = self._raw_client.login()
            except (YesWeHackRawAPiError, requests.RequestException) as e:
                raise YesWeHackApiClientError("Unable to log in with YesWeHack API client") from e
            self._logged_in = success

    def get_program_reports(
        self,
        slug: str,
        filters: Optional[Dict[str, Any]] = None,
    ) -> List[Report]:
        """
        Get reports for the program.

        Args:
            slug: a program slug
            filters: a report filter

        Returns:
            the list of reports

        Raises:
            YesWeHackApiClientError: if an error occurred when getting the reports
        """
        self._ensure_login()
        try:
            raw_reports = self._raw_client.get_reports(
                program=slug,
                filters=filters or {},
                lazy=True,
            )
        except (YesWeHackRawAPiError, YesWeHackApiClientError, requests.RequestException) as e:
            raise YesWeHackApiClientError(f"Unable to get reports for program {slug}") from e
        return self._get_detailed_reports(
            raw_reports=raw_reports,
        )

    def _get_detailed_reports(
        self,
        raw_reports: List[YesWeHackRawApiReport],
    ) -> List[Report]:
        return [
            self._get_detailed_report(
                report_id=raw_report.id,
            )
            for raw_report in raw_reports
        ]

    def _get_detailed_report(
        self,
        report_id: int,
    ) -> Report:
        try:
            raw_report = self._raw_client.get_report(
                report=report_id,
                lazy=True,
            )
        except (YesWeHackRawAPiError, requests.RequestException, TypeError) as e:
            raise YesWeHackApiClientError(f"Unable to get report #{report_id} details") from e
        return self._map_raw_report(
            raw_report=raw_report,
        )

    def _get_mapping_context(
        self,
    ) -> MappingContext:
        return MappingContext(
            yeswehack_domain=self._yeswehack_domain,
        )

    def _map_raw_report(
        self,
        raw_report: YesWeHackRawApiReport,
    ) -> Report:
        try:
            return map_raw_report(
                context=self._get_mapping_context(),
                raw_report=raw_report,
            )
        except TypeError as e:
            raise YesWeHackApiClientError(f"Unable to map report #{raw_report.id}") from e

    def _map_raw_attachment(
        self,
        raw_attachment: YesWeHackRawApiAttachment,
    ) -> Attachment:
        try:
            return map_raw_attachment(
                context=self._get_mapping_context(),
                raw_attachment=raw_attachment,
            )
        except TypeError as e:
            raise YesWeHackApiClientError(f"Unable to map attachment #{raw_attachment}") from e

    def _map_raw_logs(
        self,
        raw_logs: List[YesWeHackRawApiLog],
    ) -> List[Log]:
        try:
            return map_raw_logs(
                context=self._get_mapping_context(),
                raw_logs=raw_logs,
            )
        except TypeError as e:
            raise YesWeHackApiClientError(f"Unable to map logs #{raw_logs}") from e

    def put_report_tracking_status(
        self,
        report: Report,
        status: str,
        tracker_name: str,
        issue_id: str,
        issue_url: str,
        comment: str,
    ) -> None:
        """
        Update a report tracking status with information about a tracked issue.

        Args:
            report: a report
            status: a status (T, AFI, ...)
            tracker_name: a tracker name
            issue_id: an id of an issue from a tracker
            issue_url: an URL of an issue from a tracker
            comment: a comment

        Raises:
            YesWeHackApiClientError: if an error occurred when updating the status
        """
        self._ensure_login()
        try:
            response: requests.Response = report.raw_report.put_tracking_status(
                tracking_status=status,
                tracker_name=tracker_name,
                tracker_id=issue_id,
                tracker_url=issue_url,
                message=comment,
            )
        except (YesWeHackRawAPiError, requests.RequestException) as api_error:
            raise YesWeHackApiClientError(f"Unable to update report #{report.report_id} tracking status") from api_error
        try:
            data = response.json()
        except JSONDecodeError as decode_error:
            raise YesWeHackApiClientError(
                f"Unable to parse response after updating report #{report.report_id} tracking status",
            ) from decode_error
        if not isinstance(data, dict):
            raise YesWeHackApiClientError(
                (
                    f"Expecting {dict} from response to report #{report.report_id} tracking status update ; "
                    + f"got {type(data)}",
                ),
            )
        if not response.ok or "errors" in data:
            message = data["message"] if "message" in data and data["message"] else "[no error message]"
            raise YesWeHackApiClientError(
                f"Unable to update report #{report.report_id} tracking status: {message}",
            )

    def post_report_tracker_update(
        self,
        report: Report,
        tracker_name: str,
        issue_id: str,
        issue_url: str,
        token: str,
        comment: str,
    ) -> None:
        """
        Send a report tracker update with information about a tracked issue.

        Args:
            report: a report
            tracker_name: a tracker name
            issue_id: an id of an issue from a tracker
            issue_url: an URL of an issue from a tracker
            token: a token containing the state of the issue
            comment: a comment

        Raises:
            YesWeHackApiClientError: if an error occurred when sending the update
        """
        self._ensure_login()
        try:
            response = report.raw_report.post_tracker_update(
                tracker_name=tracker_name,
                tracker_id=issue_id,
                tracker_url=issue_url,
                token=token,
                message=comment,
            )
        except (YesWeHackRawAPiError, requests.RequestException) as api_error:
            raise YesWeHackApiClientError(f"Unable to send report #{report.report_id} tracker update") from api_error
        try:
            data = response.json()
        except JSONDecodeError as decode_error:
            raise YesWeHackApiClientError(
                f"Unable to parse response after report #{report.report_id} tracker update",
            ) from decode_error
        if not isinstance(data, dict):
            raise YesWeHackApiClientError(
                (
                    f"Expecting {dict} from response to report #{report.report_id} tracker update ; "
                    + f"got {type(data)}",
                ),
            )
        if not response.ok or "errors" in data:
            message = data["message"] if "message" in data and data["message"] else "[no error message]"
            raise YesWeHackApiClientError(
                f"Unable to update report #{report.report_id} tracker: {message}",
            )

    def post_report_attachment(
        self,
        report: Report,
        filename: str,
        file_content: bytes,
        file_type: Optional[str] = None,
    ) -> Attachment:
        """
        Send an attachment to a report.

        Args:
            report: a report
            filename: a file name
            file_content: file content
            file_type: an optional file type (mime)

        Returns:
            the attachment

        Raises:
            YesWeHackApiClientError: if an error occurred when sending the attachment
        """
        self._ensure_login()
        try:
            attachment = report.raw_report.post_attachment(
                filename=filename,
                file_content=file_content,
                file_type=file_type,
            )
        except (YesWeHackRawAPiError, requests.RequestException) as api_error:
            raise YesWeHackApiClientError(
                f"Unable to send attachment {filename} to report #{report.report_id}",
            ) from api_error
        return self._map_raw_attachment(
            raw_attachment=attachment,
        )

    def post_report_tracker_message(
        self,
        report: Report,
        tracker_name: str,
        issue_id: str,
        issue_url: str,
        comment: str,
        attachments: Optional[List[str]] = None,
    ) -> None:
        """
        Send a report tracker message with information about a tracked issue.

        Args:
            report: a report
            tracker_name: a tracker name
            issue_id: an id of an issue from a tracker
            issue_url: an URL of an issue from a tracker
            comment: a comment
            attachments: a list of attachment identifiers

        Raises:
            YesWeHackApiClientError: if an error occurred when sending the message
        """
        self._ensure_login()
        try:
            report.raw_report.post_tracker_message(
                tracker_name=tracker_name,
                tracker_id=issue_id,
                tracker_url=issue_url,
                message=comment,
                attachments=attachments,
            )
        except (YesWeHackRawAPiError, requests.RequestException) as api_error:
            raise YesWeHackApiClientError(f"Unable to send report #{report.report_id} tracker message") from api_error

    def put_status(
        self,
        report: Report,
        status: str,
        comment: str,
        attachments: Optional[List[str]] = None,
    ) -> List[Log]:
        """
        Update a report status.

        Args:
            report: a report
            status: a status
            comment: a comment
            attachments: a list of attachment identifiers

        Returns:
            a list of logs

        Raises:
            YesWeHackApiClientError: if an error occurred when updating the status
        """
        self._ensure_login()
        try:
            raw_logs = report.raw_report.put_status(
                status=status,
                message=comment,
                attachments=attachments,
            )
        except (YesWeHackRawAPiError, requests.RequestException) as api_error:
            raise YesWeHackApiClientError(f"Unable to update report #{report.report_id} status") from api_error
        return self._map_raw_logs(
            raw_logs=raw_logs,
        )

    def put_ask_for_fix_verification_status(
        self,
        report: Report,
        status: str,
        comment: str,
        attachments: Optional[List[str]] = None,
    ) -> List[Log]:
        """
        Update a report status.

        Args:
            report: a report
            status: a status
            comment: a comment
            attachments: a list of attachment identifiers

        Returns:
            a list of logs

        Raises:
            YesWeHackApiClientError: if an error occurred when updating the status
        """
        self._ensure_login()
        try:
            raw_logs = report.raw_report.put_ask_for_fix_verification_status(
                status=status,
                message=comment,
                attachments=attachments,
            )
        except (YesWeHackRawAPiError, requests.RequestException) as api_error:
            raise YesWeHackApiClientError(
                f"Unable to update report #{report.report_id} ask for fix verification status"
            ) from api_error
        return self._map_raw_logs(
            raw_logs=raw_logs,
        )



"""Models and functions used for attributes container dict entry GUI."""
from typing import (
    Any,
    Optional,
)

from PySide6.QtCore import Signal
from PySide6.QtWidgets import (
    QFormLayout,
    QLabel,
    QLineEdit,
    QVBoxLayout,
    QWidget,
)

from ywh2bt.core.configuration.attribute import AttributesContainer
from ywh2bt.gui.widgets.attribute.attributes_container_dict_entry import AttributesContainerDictEntry
from ywh2bt.gui.widgets.attribute.attributes_container_widget import AttributesContainerWidget
from ywh2bt.gui.widgets.typing import as_signal_instance


class AttributesContainerDictEntryWidget(QWidget):
    """Attributes container dict entry GUI."""

    key_changed: Signal = Signal(AttributesContainerDictEntry)
    value_changed: Signal = Signal(AttributesContainerDictEntry)

    _entry: AttributesContainerDictEntry
    _key_line_edit: QLineEdit
    _value_widget: AttributesContainerWidget

    def __init__(
        self,
        entry: AttributesContainerDictEntry,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            entry: an entry
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._entry = entry
        self._init_ui()

    def _init_ui(
        self,
    ) -> None:
        layout = QVBoxLayout(self)

        form_layout = QFormLayout()
        form_layout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)

        key_label = QLabel(
            "Key:",
            self,
        )
        self._key_line_edit = self._create_key_widget()
        self._key_line_edit.setText(self._entry.key)
        self._value_widget = self._create_value_widget()

        form_layout.addRow(key_label, self._key_line_edit)
        form_layout.addRow(self._value_widget)

        layout.addLayout(form_layout, 1)
        layout.addStretch(1)

    def _create_key_widget(
        self,
    ) -> QLineEdit:
        widget = QLineEdit(self)
        as_signal_instance(widget.textEdited).connect(
            self._on_key_changed,
        )
        return widget

    def _on_key_changed(
        self,
        value: str,
    ) -> None:
        self._entry.key = value
        as_signal_instance(self.key_changed).emit(
            self._entry,
        )

    def _create_value_widget(
        self,
    ) -> AttributesContainerWidget:
        widget = AttributesContainerWidget(
            parent=self,
            container_class=self._entry.value.__class__,
            container=self._entry.value,
        )
        as_signal_instance(widget.container_changed).connect(
            self._on_value_changed,
        )
        return widget

    def _on_value_changed(
        self,
        value: AttributesContainer,
    ) -> None:
        self._entry.value = value
        as_signal_instance(self.value_changed).emit(
            self._entry,
        )



"""Models and functions used for loading/writing configuration files."""
from dataclasses import dataclass
from typing import (
    Dict,
    OrderedDict,
    Set,
    TextIO,
    Type,
)

from ywh2bt.core.configuration.error import BaseAttributeError
from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.exceptions import CoreException
from ywh2bt.core.loader import RootConfigurationLoader
from ywh2bt.core.schema.json import root_configuration_as_json_schema
from ywh2bt.core.schema.markdown import root_configuration_as_markdown
from ywh2bt.core.schema.protocol import SchemaDumpProtocol
from ywh2bt.core.schema.text import root_configuration_as_text
from ywh2bt.core.serde import SerDe
from ywh2bt.core.serializers.json import JsonSerDe
from ywh2bt.core.serializers.yaml import YamlSerDe


@dataclass
class AvailableFormat:
    """A class describing a core format and its supported file name extensions."""

    ser_de: Type[SerDe]
    extensions: Set[str]


AVAILABLE_FORMATS: Dict[str, AvailableFormat] = OrderedDict[str, AvailableFormat](
    {
        "yaml": AvailableFormat(
            ser_de=YamlSerDe,
            extensions={
                "yaml",
                "yml",
            },
        ),
        "json": AvailableFormat(
            ser_de=JsonSerDe,
            extensions={
                "json",
            },
        ),
    }
)

AVAILABLE_SCHEMA_DUMP_FORMATS: Dict[str, SchemaDumpProtocol] = OrderedDict[str, SchemaDumpProtocol](
    {
        "text": root_configuration_as_text,
        "markdown": root_configuration_as_markdown,
        "json": root_configuration_as_json_schema,
    }
)


def get_root_configuration_loader(
    file_format: str,
) -> RootConfigurationLoader:
    """
    Get a configuration loader for the given format.

    Args:
        file_format: a file format

    Returns:
        a configuration loader
    """
    serde = get_serde(
        file_format=file_format,
    )
    return RootConfigurationLoader(
        serde=serde,
    )


def get_serde(
    file_format: str,
) -> SerDe:
    """
    Get a serde for the given format.

    Args:
        file_format: a file format

    Returns:
        a serde

    Raises:
        CoreException: is the format is not supported
    """
    available_format = AVAILABLE_FORMATS.get(file_format)
    if available_format:
        return available_format.ser_de()
    raise CoreException(f"Unsupported file format {file_format}.")


def load_configuration(
    config_format: str,
    config_file: str,
) -> RootConfiguration:
    """
    Load a configuration file.

    Args:
        config_format: a file format
        config_file: a file

    Returns:
        a configuration

    Raises:
        CoreException: if the configuration couldn't be loaded
    """
    loader = get_root_configuration_loader(
        file_format=config_format,
    )
    try:
        with open(config_file, "r") as f:
            return loader.load(
                stream=f,
            )
    except FileNotFoundError as e:
        raise CoreException("Configuration file not found") from e
    except BaseAttributeError as e:
        raise CoreException("Configuration invalid") from e


def write_message(
    stream: TextIO,
    message: str,
    end: str = "\n",
) -> None:
    """
    Write a message on the stream.

    Args:
        stream: a stream
        message: a message
        end: a string appended after the message, default a newline
    """
    stream.write(message)
    if end:
        stream.write(end)
    stream.flush()



"""Misc. constants used in the GUI."""
from PySide6.QtCore import QSize


SMALL_BUTTON_WIDTH = 24
SMALL_BUTTON_HEIGHT = 20
SMALL_BUTTON_SIZE = QSize(
    SMALL_BUTTON_WIDTH,
    SMALL_BUTTON_HEIGHT,
)

SMALL_BUTTON_ICON_WIDTH = 15
SMALL_BUTTON_ICON_HEIGHT = 15
SMALL_BUTTON_ICON_SIZE = QSize(
    SMALL_BUTTON_ICON_WIDTH,
    SMALL_BUTTON_ICON_HEIGHT,
)

BIG_BUTTON_WIDTH = 120
BIG_BUTTON_HEIGHT = 60
BIG_BUTTON_SIZE = QSize(
    BIG_BUTTON_WIDTH,
    BIG_BUTTON_HEIGHT,
)

BIG_BUTTON_ICON_WIDTH = 32
BIG_BUTTON_ICON_HEIGHT = 32
BIG_BUTTON_ICON_SIZE = QSize(
    BIG_BUTTON_ICON_WIDTH,
    BIG_BUTTON_ICON_HEIGHT,
)

TAB_WIDGET_CORNER_BUTTON_WIDTH = 40
TAB_WIDGET_CORNER_BUTTON_HEIGHT = 20
TAB_WIDGET_CORNER_BUTTON_SIZE = QSize(
    TAB_WIDGET_CORNER_BUTTON_WIDTH,
    TAB_WIDGET_CORNER_BUTTON_HEIGHT,
)



"""Models and classes related to validatable objects."""
from abc import (
    ABC,
    abstractmethod,
)

from ywh2bt.core.exceptions import CoreException


class Validatable(ABC):
    """A class representing an object that can be validated."""

    @abstractmethod
    def validate(self) -> None:
        """Validate the object."""

    def is_valid(self) -> bool:
        """
        Check if the object is valid.

        Returns:
            a boolean indicating if the object is valid
        """
        try:
            self.validate()
        except CoreException:
            return False
        return True



"""Models and functions used for states decryption."""
import base64
import json
import re
from typing import (
    Optional,
    Type,
    TypeVar,
)

from ywh2bt.core.crypt.decrypt import Decryptor
from ywh2bt.core.crypt.error import CryptError
from ywh2bt.core.crypt.key import Key
from ywh2bt.core.state.error import StateError
from ywh2bt.core.state.state import State


StateType = TypeVar("StateType", bound=State, covariant=True)


class StateDecryptor:
    """A simple state decryptor."""

    _encrypted_re = re.compile(
        r"\[YWH2BT:S:(.+)]",
    )

    @classmethod
    def decrypt(
        cls,
        encrypted_state: str,
        key: str,
        state_type: Type[StateType],
    ) -> Optional[StateType]:
        """
        Decrypt a state with a key.

        Args:
            encrypted_state: an encrypted state
            key: a cryptographic key
            state_type: a type of the decrypted state

        Raises:
            StateError: if the state could not be decrypted

        Returns:
            The decrypted state.
        """
        encrypted_base64 = cls._extract_encrypted(
            encrypted_state=encrypted_state,
        )
        if not encrypted_base64:
            return None
        encrypted = base64.b64decode(encrypted_base64)
        decryptor = Decryptor(
            key=Key.from_str(
                key=key,
            ),
        )
        try:
            decrypted_data = decryptor.decrypt(encrypted)
        except CryptError as decrypt_error:
            raise StateError(f"Unable to decrypt state {encrypted_state}") from decrypt_error
        try:
            class_name, data = json.loads(decrypted_data)
        except (json.JSONDecodeError, TypeError) as json_error:
            raise StateError(f"Unable to deserialize state {encrypted_state}") from json_error
        if class_name != state_type.__name__:
            raise StateError(f"Invalid state type {class_name}")
        if not isinstance(data, dict):
            raise StateError("Invalid state data")
        return state_type(**data)

    @classmethod
    def _extract_encrypted(
        cls,
        encrypted_state: str,
    ) -> Optional[str]:
        match = cls._encrypted_re.search(encrypted_state)
        if match:
            return match.group(1)
        return None



"""Models used to describe bugtrackers."""
from __future__ import annotations

import os
import pkgutil
import sys
from importlib import import_module
from typing import (
    Any,
    Dict,
)

from ywh2bt.core.configuration.attribute import (
    AttributesContainer,
    AttributesContainerDict,
)
from ywh2bt.core.configuration.subtypable import (
    Subtypable,
    SubtypableMetaclass,
    SubtypeError,
)


class TrackerConfiguration(AttributesContainer, Subtypable, metaclass=SubtypableMetaclass):
    """Base class for tracker."""

    def __init__(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the tracker.

        Args:
            kwargs: keyword arguments

        Raises:
            SubtypeError: if the tracker has not been registered
        """
        tracker_type = self.__class__.get_subtype_name(self.__class__)
        if tracker_type is None:
            message = (
                f"TrackerConfiguration {self.__class__} is not registered ; "
                + "use TrackerConfiguration.register_subtype() to register."
            )
            raise SubtypeError(
                message=message,
                context=self,
            )
        super().__init__(**kwargs)

    def export(self) -> Dict[str, Any]:
        """
        Export the tracker data, adding the registered tracker type as `type` entry.

        Returns:
            The data.
        """
        exported = {}
        tracker_type = self.__class__.get_subtype_name(self.__class__)
        if tracker_type:
            exported["type"] = tracker_type
        exported.update(super().export())
        return exported


# import all defined trackers
module_name = vars(sys.modules[__name__])["__package__"]
dirname = os.path.dirname(__file__)
for (_module_finder, name, _is_pkg) in pkgutil.iter_modules([f"{dirname}/trackers"]):
    import_module(f"{module_name}.trackers.{name}")


class Trackers(AttributesContainerDict[TrackerConfiguration]):
    """A list of trackers."""

    def __init__(
        self,
        **kwargs: Any,
    ):
        """
        Initialize the list of trackers.

        Args:
            kwargs: a list of trackers
        """
        super().__init__(
            values_type=TrackerConfiguration,
            items=kwargs,
        )



"""Models used for the configuration of a GitLab tracker."""
from typing import (
    Any,
    Optional,
    Text,
)

from ywh2bt.core.configuration.attribute import (
    Attribute,
    BoolAttributeType,
    StrAttributeType,
)
from ywh2bt.core.configuration.tracker import TrackerConfiguration
from ywh2bt.core.configuration.validator import (
    not_blank_validator,
    url_validator,
)


class GitLabConfiguration(TrackerConfiguration):
    """A class describing the configuration of a GitLab tracker."""

    url: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="API URL",
        description="Base URL of the GitLab server",
        default="https://gitlab.com",
        validator=url_validator,
    )
    token: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="API token",
        description="User private token for the GitLab API",
        required=True,
        secret=True,
        validator=not_blank_validator,
    )
    project: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="Project path",
        description="Path to the project on GitLab",
        required=True,
        validator=not_blank_validator,
    )
    verify: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Verify TLS",
        description="Verify server's TLS certificate",
        default=True,
    )
    confidential: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Confidential issues",
        description="Mark created issues as confidential",
        default=False,
    )

    def __init__(
        self,
        url: Optional[Text] = None,
        token: Optional[Text] = None,
        project: Optional[Text] = None,
        verify: Optional[bool] = None,
        confidential: Optional[bool] = None,
        **kwargs: Any,
    ):
        """
        Initialize the configuration.

        Args:
            url: a GitLab API URL
            token: a GitLab API token
            project: a GitLab project name
            verify: a flag indicating whether to check SSL/TLS connection
            confidential: a flag indicating whether to mark created issues as confidential
            kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.url = url
        self.token = token
        self.project = project
        self.verify = verify
        self.confidential = confidential


TrackerConfiguration.register_subtype(
    subtype_name="gitlab",
    subtype_class=GitLabConfiguration,
)



"""Models and functions used for root configuration GUI."""
from __future__ import annotations

from dataclasses import dataclass
from io import StringIO
from typing import Optional

from PySide6.QtCore import (
    QByteArray,
    QFileInfo,
)

from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.core import get_root_configuration_loader
from ywh2bt.core.exceptions import CoreException
from ywh2bt.core.loader import LoaderError
from ywh2bt.gui.hashing import file_checksum


@dataclass
class RootConfigurationEntryFile:
    """A class describing an exiting configuration file."""

    info: QFileInfo
    file_format: str
    checksum: QByteArray


class RootConfigurationEntry:
    """A class describing an editable configuration."""

    _name: str
    _configuration: Optional[RootConfiguration]
    _raw: str
    _original_raw: str
    _raw_format: str
    _file: Optional[RootConfigurationEntryFile]

    def __init__(
        self,
        name: str,
        raw: str,
        original_raw: str,
        raw_format: str,
        file: Optional[RootConfigurationEntryFile] = None,
    ):
        """
        Initialize self.

        Args:
            name: a configuration name
            raw: raw content of the configuration
            original_raw: original raw content of the configuration
            raw_format: a format (json, yaml, ...)
            file: associated file information
        """
        self._name = name
        self._raw = raw
        self._original_raw = original_raw
        self._raw_format = raw_format
        self._file = file
        self._configuration = self._configuration_from_raw(
            raw=self._raw,
            raw_format=self._raw_format,
        )

    def _update_configuration(
        self,
    ) -> None:
        self._configuration = None
        self._configuration = self._configuration_from_raw(
            raw=self._raw,
            raw_format=self._raw_format,
        )

    def _update_raw(
        self,
    ) -> None:
        self._raw = self._raw_from_configuration()

    @property
    def name(
        self,
    ) -> str:
        """
        Get the name.

        Returns:
            name
        """
        return self._name

    @name.setter
    def name(
        self,
        name: str,
    ) -> None:
        """
        Set the name.

        Args:
            name: a name
        """
        self._name = name

    @property
    def configuration(
        self,
    ) -> Optional[RootConfiguration]:
        """
        Get the configuration.

        Returns:
            configuration
        """
        return self._configuration

    @configuration.setter
    def configuration(
        self,
        configuration: Optional[RootConfiguration],
    ) -> None:
        """
        Set the configuration.

        Args:
            configuration: a configuration
        """
        self._configuration = configuration
        self._update_raw()

    @property
    def original_configuration(
        self,
    ) -> Optional[RootConfiguration]:
        """
        Get the original configuration.

        Returns:
            original configuration
        """
        return self._configuration_from_raw(
            raw=self._original_raw,
            raw_format=self._raw_format,
        )

    @property
    def raw(
        self,
    ) -> str:
        """
        Get the raw content.

        Returns:
            raw content
        """
        return self._raw

    @raw.setter
    def raw(
        self,
        raw: str,
    ) -> None:
        """
        Set the raw content.

        Args:
            raw: content
        """
        self._raw = raw
        self._update_configuration()

    @property
    def raw_format(
        self,
    ) -> str:
        """
        Get the raw format.

        Returns:
            raw format
        """
        return self._raw_format

    @raw_format.setter
    def raw_format(
        self,
        raw_format: str,
    ) -> None:
        """
        Set the raw format.

        Args:
            raw_format: a format
        """
        self._raw_format = raw_format

    @property
    def original_raw(
        self,
    ) -> str:
        """
        Get the original raw content.

        Returns:
            original raw content
        """
        return self._original_raw

    @original_raw.setter
    def original_raw(
        self,
        original_raw: str,
    ) -> None:
        """
        Set the original raw content.

        Args:
            original_raw: content
        """
        self._original_raw = original_raw

    @property
    def file(
        self,
    ) -> Optional[RootConfigurationEntryFile]:
        """
        Get the file information.

        Returns:
            file information
        """
        return self._file

    @file.setter
    def file(
        self,
        file: Optional[RootConfigurationEntryFile],
    ) -> None:
        """
        Set the file information.

        Args:
            file: file information
        """
        self._file = file

    def has_changed(self) -> bool:
        """
        Check if the raw content has changed.

        Returns:
            True if the file has changed, otherwise False.
        """
        return self._raw != self._original_raw

    def is_empty(self) -> bool:
        """
        Check if the configuration is empty.

        Returns:
            True if the configuration is empty, otherwise False.
        """
        return self._raw == ""

    @classmethod
    def from_file(
        cls,
        file_info: QFileInfo,
        file_format: str,
    ) -> RootConfigurationEntry:
        """
        Create a RootConfigurationEntry from a file.

        Args:
            file_info: a file
            file_format: a file format

        Raises:
            CoreException: if the file couldn't be opened

        Returns:
            a RootConfigurationEntry
        """
        file_path = file_info.filePath()
        try:
            with open(file_path, "r") as f:
                raw = f.read()
        except IOError as e:
            raise CoreException(f"Unable to open file:\n{file_path}") from e

        return RootConfigurationEntry(
            name=file_info.fileName(),
            raw=raw,
            original_raw=raw,
            raw_format=file_format,
            file=RootConfigurationEntryFile(
                info=file_info,
                file_format=file_format,
                checksum=file_checksum(file_info.filePath()),
            ),
        )

    @classmethod
    def _configuration_from_raw(
        cls,
        raw: str,
        raw_format: str,
    ) -> Optional[RootConfiguration]:
        if not raw:
            return RootConfiguration()
        loader = get_root_configuration_loader(
            file_format=raw_format,
        )
        raw_stream = StringIO(raw)
        try:
            return loader.load(
                stream=raw_stream,
            )
        except LoaderError as e:
            raise CoreException("Invalid raw configuration") from e

    def _raw_from_configuration(
        self,
    ) -> str:
        if not self._configuration:
            return ""
        loader = get_root_configuration_loader(
            file_format=self._raw_format,
        )
        raw_stream = StringIO()
        try:
            loader.save(
                stream=raw_stream,
                data=self._configuration,
            )
        except LoaderError:
            return ""
        return raw_stream.getvalue()
